#include <gint/display.h>
#include "extrakeyboard.h"
#include "utilities.h"

#include <libprof.h>
#include <fxlibc/printf.h>

#include <stdbool.h>


extern bopti_image_t img_player;
extern bopti_image_t img_rock;


bool exit = false;
KeyboardExtra MyKeyboard;


int selector = 0;
int16_t xmin[2], ymin[2];
int16_t xmax[2], ymax[2];
SpriteLocator sprite1, sprite2;


static void get_inputs( )
{
    MyKeyboard.Update( 0.0f );

	if(MyKeyboard.IsKeyPressedEvent(MYKEY_F1)) 	  (selector = 0);
	if(MyKeyboard.IsKeyPressedEvent(MYKEY_F2)) 	  (selector = 1);

	if(MyKeyboard.IsKeyPressed(MYKEY_EXIT)) {exit = true; };

    if(MyKeyboard.IsKeyPressed(MYKEY_LEFT))   { if(xmin[selector]>0) xmin[selector]--;}
    if(MyKeyboard.IsKeyPressed(MYKEY_RIGHT))  { if(xmax[selector]<DWIDTH) xmin[selector]++;}
    if(MyKeyboard.IsKeyPressed(MYKEY_UP))     { if(ymin[selector]>0) ymin[selector]--;}
    if(MyKeyboard.IsKeyPressed(MYKEY_DOWN))   { if(ymax[selector]<DHEIGHT-50) ymin[selector]++;}
}


void update()
{
	xmax[0] = xmin[0] + img_player.width;
	ymax[0] = ymin[0] + img_player.height;

	xmax[1] = xmin[1] + img_rock.width;
	ymax[1] = ymin[1] + img_rock.height;

	sprite1 = { .x = xmin[0], .y = ymin[0],	.image = &img_player };
	sprite2 = { .x = xmin[1], .y = ymin[1], .image = &img_rock };	
}


int main(void)
{
	/* These are declaration of stuff to be used with LibProfg to measure performances */
	uint32_t time_AABB=0, time_FastPP=0, time_DebugPP=0;
	prof_t perf_AABB, perf_FastPP, perf_DebugPP;

	prof_init();

    __printf_enable_fp();


	/* initial declaration of images position and data */
	xmin[0] = 10;
	ymin[0] = 10;
	xmin[1] = DWIDTH/2 - img_rock.width/2;
	ymin[1] = (DHEIGHT-50)/2 - img_rock.height/2;
	sprite1 = { .x = xmin[0], .y = ymin[0], .image = &img_player };
	sprite2 = { .x = xmin[1], .y = ymin[1], .image = &img_rock };	
	update();


	/* boolean used to track if we have collsion or not (one for each method) */
	bool AABB = false;
	bool FastPP = false;
	bool DebugPP = false;


	do
	{

		get_inputs();
		update();


		dclear( C_BLACK );

		/* drawx the sprites wih surrounding boxes */
		dimage( xmin[0], ymin[0], &img_player );
		drect_border( xmin[0], ymin[0], xmax[0], ymax[0], C_NONE, 1, C_WHITE );

		dimage( xmin[1], ymin[1], &img_rock );
		drect_border( xmin[1], ymin[1], xmax[1], ymax[1], C_NONE, 1, C_WHITE );

		/* AABB collision test */
		perf_AABB = prof_make();
        prof_enter(perf_AABB);

			AABB = AABB_Collision( sprite1, sprite2 );

		prof_leave(perf_AABB);
        time_AABB = prof_time(perf_AABB);

		/* Pixel Perfect collision test */
		perf_FastPP = prof_make();
        prof_enter(perf_FastPP);

			FastPP = Pixel_Perfect_Collision( sprite1, sprite2 );

		prof_leave(perf_FastPP);
        time_FastPP = prof_time(perf_FastPP);

		/* Pixel Perfect collision test but with internal rendering routines (Very slow, but needed to check what's happening internally)*/
		perf_DebugPP = prof_make();
        prof_enter(perf_DebugPP);

			DebugPP = DEBUG_Pixel_Perfect_Collision( sprite1, sprite2 );

		prof_leave(perf_DebugPP);
        time_DebugPP = prof_time(perf_DebugPP);


		/* Visual restitution of tests (GREEN = Not Colliding / RED = Colliding) */
		/* Times are all in µs (microseconds = 1/1000000 sec) */
		dtext( 50, DHEIGHT-35, C_WHITE, "AABB");
		if (AABB) drect( 10, DHEIGHT-40, 40, DHEIGHT-10, C_RED);
		else drect( 10, DHEIGHT-40, 40, DHEIGHT-10, C_GREEN );
		dprint( 10, DHEIGHT-35, C_WHITE, "%d", time_AABB);


		dtext( 215, DHEIGHT-35, C_WHITE, "   Pixel Perfect");
		dtext( 215, DHEIGHT-20, C_WHITE, "<DEBUG   vs   Fast>");
		if (FastPP) drect( DWIDTH-40, DHEIGHT-40, DWIDTH-10, DHEIGHT-10, C_RED);
		else drect( DWIDTH-40, DHEIGHT-40, DWIDTH-10, DHEIGHT-10, C_GREEN );
		dprint( DWIDTH-40, DHEIGHT-35, C_WHITE, "%d", time_FastPP);



		if (DebugPP) drect( DWIDTH/2-15, DHEIGHT-40, DWIDTH/2+15, DHEIGHT-10, C_RED);
		else drect( DWIDTH/2-15, DHEIGHT-40, DWIDTH/2+15, DHEIGHT-10, C_GREEN );
		dprint( DWIDTH/2-15, DHEIGHT-35, C_WHITE, "%d", time_DebugPP);


		dupdate();
    }
    while (exit==false);

    prof_quit();

	return 1;
}
